
--
--  TjDropDownMenu.lua
--    by Tuhljin
--
--  Alternative drop down menu system. A simple way to implement dropdown menus where one and only one thing should
--  be checked. Supports nested menus as well (but again, only one thing can be checked across the entire menu).
--  (Instructions are lacking in this early version, sorry. Peruse the comments of this file and see the Gossipmonger
--  addon for example usage (though it doesn't utilize all of TjDDM's features). You should be familiar with defining
--  tables for use with functions such as the Blizzard-provided EasyMenu if you are to use this system.)
--
--  Use "NO_TjDDM_Setup = 1" (w/o quotes) in menu table entries to prevent that entry from being touched even if you
--  want other parts of the table automatically altered.
--
--  There are a few problems with using the menu properties notClickable and notCheckable (not due to TjDDM, but in
--  general), so an alternative for those items you don't want to be checked is presented: TjDDM_notCheckable. (Set
--  to 1 in the menu table entry to use.) It can be used as you see fit, but the primary intent was for it to be
--  paired with keepShownOnClick for those entries that hold a sub-menu via the menuList property.
--
--  Be sure that all checkable entries have a "value" property that is unique for that menu (including submenus).
--  If no value is given, TjDDM will assign one, but authors will often find it more useful to use preset values.
--
--  Tooltip use: Set your dropdown's .tooltip (and optionally .tooltip2) properties to display a specific tooltip on
--  mouseover. Set .tooltip to 0 to show no tooltip. Otherwise, we use the currently selected entry's tooltip, if
--  it has one.
--


local THIS_VERSION = 0.2

if (not TjDropDownMenu or TjDropDownMenu.Version < THIS_VERSION) then
  TjDropDownMenu = {};
  TjDropDownMenu.Version = THIS_VERSION;

  local xOffset, yOffset = 9, 7              -- Base offset (where offset of 0, 0 would put the menu frame).
  local xOffset_MENU, yOffset_MENU = 17, 5   -- Base offset for displayMode "MENU"
  local generateNewMenuFrameForAll = true;  -- Set to true to generate a new menuFrame object for every dropdown.
     -- Otherwise, those of the same displayMode are shared. This is a static option, to be set here in the LUA only
     -- (not meant to be changed dynamically).

  local menuFrameList
  if (not generateNewMenuFrameForAll) then
    menuFrameList = {};
  end

  local NumCreated = 0;
  local hooksComplete, TjMenuOpen, prevTjMenuOpen, clickedButton;

  local function debugprint(msg)
    DEFAULT_CHAT_FRAME:AddMessage("TjDDM: "..msg, 1, 0.7, 0.6);
  end

  local function SelectEntry(dropdown, tab)
    tab.checked = 1
    dropdown.TjDDM.selectedValue = tab.value
    dropdown:SetText(tab.text)

    local callFunc = dropdown.TjDDM.OnSelectFunc
    if (type(callFunc) == "function") then
      callFunc(dropdown, tab.value, tab);
    end
  end


-- AUTOMATIC FUNCTION SETUP FOR MENU ITEMS
----------------------------------------------

  local function RefreshMenu(frame, useValue, dropdownLevel)
  -- Largely taken from UIDropDownMenu_Refresh in UIDropDownMenu.lua, which didn't do things quite right for us.
  -- Comments in this function taken from there as well, except those starting with "TjDDM."
    local button, checked, checkImage, normalText, width;
    local maxWidth = 0;
    if ( not frame ) then
      frame = this;
    end
    if ( not dropdownLevel ) then
      dropdownLevel = UIDROPDOWNMENU_MENU_LEVEL;
    end

    -- Just redraws the existing menu
    for i=1, UIDROPDOWNMENU_MAXBUTTONS do
      button = getglobal("DropDownList"..dropdownLevel.."Button"..i);
      checked = nil;
      -- See if checked or not
      -- TjDDM: This is what we changed: It now properly knows what is checked (that is, what was just clicked).
      if ( button == clickedButton ) then
          checked = 1;
      end

      -- If checked show check image
      checkImage = getglobal("DropDownList"..dropdownLevel.."Button"..i.."Check");
      if ( checked ) then
        if ( useValue ) then
          UIDropDownMenu_SetText(button.value, frame);
        else
          UIDropDownMenu_SetText(button:GetText(), frame);
        end
        button:LockHighlight();
        checkImage:Show();
      else
        button:UnlockHighlight();
        checkImage:Hide();
      end

      if ( button:IsShown() ) then
        normalText = getglobal(button:GetName().."NormalText");
        -- Determine the maximum width of a button
        width = normalText:GetWidth() + 40;
        -- Add padding if has and expand arrow or color swatch
        if ( button.hasArrow or button.hasColorSwatch ) then
          width = width + 10;
        end
        if ( button.notCheckable ) then
          width = width - 30;
        end
        if ( width > maxWidth ) then
          maxWidth = width;
        end
      end
    end
    for i=1, UIDROPDOWNMENU_MAXBUTTONS do
      button = getglobal("DropDownList"..dropdownLevel.."Button"..i);
      button:SetWidth(maxWidth);
    end
    getglobal("DropDownList"..dropdownLevel):SetWidth(maxWidth+15);

    -- TjDDM: Added the following to recursively refresh lower levels.
    if (dropdownLevel > 1) then
      RefreshMenu(frame, useValue, dropdownLevel - 1);
    end
  end

  local function ClearChecks(menuList, includeChildren)
    for _, tab in ipairs(menuList) do
      if (type(tab) == "table" and not tab.NO_TjDDM_Setup) then
        tab.checked = nil
        if (includeChildren and type(tab.menuList) == "table") then
          ClearChecks(tab.menuList, true)
        end
      end
    end
  end

  local function menuItemClick(tab, frame, checked)
    local prevchecked = checked

    local dropdown
    if (TjMenuOpen) then
      dropdown = getglobal(TjMenuOpen)
    else
      dropdown = getglobal(prevTjMenuOpen)
    end

    if (tab.TjDDM_notCheckable) then
      checked = nil;
      tab.checked = nil;
      getglobal(clickedButton:GetName().."Check"):Hide();
    elseif (not checked or tab.keepShownOnClick) then
    -- Overrides default keepShownOnClick functionality, which toggles between checked and unchecked.
      if (TjMenuOpen and not tab.keepShownOnClick) then
      -- If the menu is still open even though this entry should close its part of the menu, it must have been in
      -- a submenu, so close the rest of it.
        CloseMenus()
      end

      ClearChecks(dropdown.TjDDM.menuList, true)
      checked = 1
      SelectEntry(dropdown, tab)
    end
    if (TjMenuOpen and not tab.TjDDM_notCheckable) then
      local menuFrame;
      if (generateNewMenuFrameForAll) then
        menuFrame = dropdown.TjDDM.menuFrame;
      else
        menuFrame = menuFrameList[dropdown.TjDDM.menuFrameType];
      end
      RefreshMenu(menuFrame)
    end

    local prevfunc = tab.TjDDM.prevfunc
    if (type(prevfunc) == "function") then
      -- Note that we pass the possibly-altered checked as the 3rd arg and the original value for it as the 4th.
      prevfunc(tab.TjDDM.prevarg1, tab.TjDDM.prevarg2, checked, prevchecked)
    end
  end

  local function SetupFunctions(dropdown, menuList, valueNum)
    valueNum = valueNum or 0
    for _, tab in ipairs(menuList) do
      if (type(tab) == "table" and not tab.NO_TjDDM_Setup and not tab.TjDDM) then
        if (not tab.isTitle and not tab.notClickable) then
          tab.TjDDM = {};
          tab.TjDDM.prevfunc = tab.func;
          tab.TjDDM.prevarg1 = tab.arg1;
          tab.TjDDM.prevarg2 = tab.arg2;
          tab.func = menuItemClick;
          tab.arg1 = tab;
          tab.arg2 = nil;
          tab.checked = nil;

          valueNum = valueNum + 1;
          if (not tab.value) then
            tab.value = "*TjDDM autovalue* "..valueNum
          end

          -- Find first valid entry for default selected value:
          if (not dropdown.TjDDM.selectedValue and not tab.TjDDM_notCheckable) then
            dropdown.TjDDM.selectedValue = tab.value;
            dropdown:SetText(tab.text)
            tab.checked = 1;
          end
        end
        -- Iterate over child menus:
        if (type(tab.menuList) == "table") then
          SetupFunctions(dropdown, tab.menuList, valueNum)
        end
      end
    end
  end


-- OTHER FUNCTIONS
----------------------------------------------

  local function findEntryByValue(menuList, value, checkNoAutoSetupEntries)
    local tab2
    for _, tab in ipairs(menuList) do
      if ( type(tab) == "table" and (checkNoAutoSetupEntries or not tab.NO_TjDDM_Setup) ) then
        if (tab.value == value) then
          return tab;
        end
        if (type(tab.menuList) == "table") then
          tab2 = findEntryByValue(tab.menuList, value, checkNoAutoSetupEntries)
          if (tab2) then  -- If recursive call found the entry, return it.
            return tab2;
          end
        end
      end
    end
  end

  local function MenuHidden()
    prevTjMenuOpen = TjMenuOpen;
    TjMenuOpen = nil;
  end

  function TjDropDownMenu.DropBtnOnClick()
    PlaySound("igMainMenuOptionCheckBoxOn");
    local dropdown = this:GetParent();
    local menuFrame;
    if (generateNewMenuFrameForAll) then
      menuFrame = dropdown.TjDDM.menuFrame;
    else
      menuFrame = menuFrameList[dropdown.TjDDM.menuFrameType];
    end

    if (not TjMenuOpen or TjMenuOpen ~= dropdown:GetName()) then
      local callFunc = dropdown.TjDDM.EasyMenuSubFunc or EasyMenu;
      -- If using another function, it should take the same arguments as EasyMenu does or problems may occur.

      local x, y = dropdown.TjDDM.x or 0, dropdown.TjDDM.y or 0;
      if (dropdown.TjDDM.displayMode == "MENU") then
        x, y = x + xOffset_MENU, y + yOffset_MENU
      else
        x, y = x + xOffset, y + yOffset
      end

      callFunc(dropdown.TjDDM.menuList, menuFrame, dropdown, x, y, dropdown.TjDDM.displayMode, dropdown.TjDDM.autoHideDelay);
      -- arg order: menuList, menuFrame, anchor, x, y, displayMode, autoHideDelay
      TjMenuOpen = dropdown:GetName();   -- Do this after callFunc or it will be set to nil by MenuHidden().
    else
      CloseMenus()
    end
  end


  local function SetOffset(frame, x, y)
    frame.TjDDM.x = x;
    frame.TjDDM.y = y;
  end

  local function GetOffset(frame, x, y)
    return frame.TjDDM.x, frame.TjDDM.y;
  end

  local function SetAutoHideDelay(frame, delay)
    frame.TjDDM.autoHideDelay = delay;
  end

  local function GetAutoHideDelay(frame)
    return frame.TjDDM.autoHideDelay;
  end

  local function SetDropDownWidth(frame, width)
    getglobal(frame:GetName().."Middle"):SetWidth(width)
    frame:SetWidth(width+8)
  end

  local function GetDropDownWidth(frame, ...)
    return getglobal(frame:GetName().."Middle"):GetWidth()
  end

  local function SetText(frame, ...)
    getglobal(frame:GetName().."Text"):SetText(...)
  end

  local function GetText(frame)
    return getglobal(frame:GetName().."Text"):GetText()
  end

  local function GetSelectedValue(frame)
    return frame.TjDDM.selectedValue;
  end

  local function SetSelectedValue(frame, value, checkNoAutoSetupEntries)
  -- Set third arg to true to include entries with NO_TjDDM_Setup properties.
    local tab = findEntryByValue(frame.TjDDM.menuList, value, checkNoAutoSetupEntries);
    if (tab) then
      ClearChecks(frame.TjDDM.menuList, true)
      SelectEntry(frame, tab)
    end
  end

  local function SetLabel(frame, text, highlight)
    if (type(text) ~= "string") then
      error("SetLabel(): Invalid argument. Expected string. Got "..type(text)..".")
    end
    if (not frame.TjDDM.label) then
      local font
      if (highlight) then  font = "GameFontHighlight";  else  font = "GameFontNormal";  end
      frame.TjDDM.label = frame:CreateFontString(frame:GetName().."Label", "ARTWORK", font);
      frame.TjDDM.label:SetPoint("BOTTOMLEFT", frame, "TOPLEFT", 16, 3);
    end
    frame.TjDDM.label:SetText(text);
  end

  local function OnSelect(frame, func)
    frame.TjDDM.OnSelectFunc = func;
  end

  local function OnEnter()
    if (this.tooltip) then
      if (this.tooltip ~= 0) then
        GameTooltip:SetOwner(this, "ANCHOR_RIGHT");
        GameTooltip:SetBackdropColor(TOOLTIP_DEFAULT_BACKGROUND_COLOR.r, TOOLTIP_DEFAULT_BACKGROUND_COLOR.g, TOOLTIP_DEFAULT_BACKGROUND_COLOR.b);
        GameTooltip:AddLine(this.tooltip, nil, nil, nil, 1);
        if (this.tooltip2) then
          GameTooltip:AddLine(this.tooltip2, nil, nil, nil, 1)
        end
        GameTooltip:Show()
      end
    else
      local tab = findEntryByValue(this.TjDDM.menuList, this.TjDDM.selectedValue, true);  -- Get current selected entry's table
      if (tab.tooltipTitle) then
        GameTooltip:SetOwner(this, "ANCHOR_RIGHT");
        GameTooltip:SetBackdropColor(TOOLTIP_DEFAULT_BACKGROUND_COLOR.r, TOOLTIP_DEFAULT_BACKGROUND_COLOR.g, TOOLTIP_DEFAULT_BACKGROUND_COLOR.b);
        GameTooltip:SetText(tab.tooltipTitle, 1, 1, 1)
        if (tab.tooltipText) then
          GameTooltip:AddLine(tab.tooltipText, NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b, 1);
          GameTooltip:Show()
        end
      end
    end
  end

  local function OnLeave()
    GameTooltip:Hide()
  end


  local orig_UIDropDownMenuButton_OnClick;
  local function hooked_UIDropDownMenuButton_OnClick(...)
    if (TjMenuOpen) then
      clickedButton = this;
    else
      clickedButton = nil;
    end
    orig_UIDropDownMenuButton_OnClick(...)
  end


  function TjDropDownMenu.CreateDropDown(name, parent, menuList, x, y, displayMode, autoHideDelay, noAutoMenuFuncSetup, EasyMenuSubFunc)
  -- Note that autoHideDelay is currently unused by EasyMenu, though it lists it as an argument.
  -- Perhaps a use for it will be implemented in the future.
    if (name and type(name) ~= "string") then
      error("TjDropDownMenu.CreateDropDown(): Invalid name (arg 1). Expected string or nil. Got "..type(name)..".")
    end

    NumCreated = NumCreated + 1;
    name = name or "TjDropDownMenu_Num"..NumCreated
    local dropdown = CreateFrame("Frame", name, parent, "UIDropDownMenuTemplate")
    dropdown.TjDDM = {};
    dropdown.TjDDM.menuList = menuList;
    dropdown.TjDDM.x = x;
    dropdown.TjDDM.y = y;
    dropdown.TjDDM.displayMode = displayMode;
    dropdown.TjDDM.autoHideDelay = autoHideDelay;
    dropdown.TjDDM.EasyMenuSubFunc = EasyMenuSubFunc;

    if (generateNewMenuFrameForAll) then
      dropdown.TjDDM.menuFrame = CreateFrame("Frame", name.."_TjDropDownMenuFrame", parent, "UIDropDownMenuTemplate")
    else
      displayMode = displayMode or 0
      dropdown.TjDDM.menuFrameType = displayMode
      if (not menuFrameList[displayMode]) then
        menuFrameList[displayMode] = CreateFrame("Frame", "TjDropDownMenu_SharedMenuFrame_"..displayMode, parent, "UIDropDownMenuTemplate")
      end
    end

    getglobal(name.."Button"):SetScript("OnClick", TjDropDownMenu.DropBtnOnClick)
    dropdown:SetScript("OnEnter", OnEnter)
    dropdown:SetScript("OnLeave", OnLeave)
    dropdown:EnableMouse(true)

    -- Set functions to call in this style: MyDropDownFrame:FunctionName(...)
    dropdown.SetOffset = SetOffset;
    dropdown.GetOffset = GetOffset;
    dropdown.SetAutoHideDelay = SetAutoHideDelay;
    dropdown.GetAutoHideDelay = GetAutoHideDelay;
    dropdown.SetDropDownWidth = SetDropDownWidth;
    dropdown.GetDropDownWidth = GetDropDownWidth;
    dropdown.SetText = SetText;
    dropdown.GetText = GetText;
    dropdown.GetSelectedValue = GetSelectedValue;
    dropdown.SetSelectedValue = SetSelectedValue;
    dropdown.OnSelect = OnSelect;
    dropdown.SetLabel = SetLabel;

    if (not noAutoMenuFuncSetup) then
      SetupFunctions(dropdown, menuList)
    end

    if (not hooksComplete) then
      DropDownList1:HookScript("OnHide", MenuHidden);
      orig_UIDropDownMenuButton_OnClick = UIDropDownMenuButton_OnClick;
      UIDropDownMenuButton_OnClick = hooked_UIDropDownMenuButton_OnClick;
      hooksComplete = true
    end

    return dropdown;
  end

end
